import tkinter as tk
from tkinter import ttk
from tkinter import messagebox
import configparser
import webbrowser
import datetime
from utils import *

class LabeledEntry(ttk.Frame):
    def __init__(self, parent, labelText, entryText=""):
        super().__init__(master=parent)
        self.label = ttk.Label(self, text=labelText)
        self.entry = ttk.Entry(self)
        self.entry.insert(0, entryText)
        
        self.label.pack(padx=(0,5), side='left', expand=False, anchor='w')
        self.entry.pack(padx=(5,0), side='left', expand=True, fill="x")
        
        
class LabeledEntryButton(ttk.Frame):
    def __init__(self, parent, labelText, buttonText, entryText=""):
        super().__init__(master=parent)
        self.label = ttk.Label(self, text=labelText)
        self.entry = ttk.Entry(self)
        self.entry.insert(0, entryText)
        self.button = ttk.Button(self, text=buttonText)
        
        self.label.pack(padx=(0,5), side='left', expand=False)
        self.entry.pack(padx=(5,5), side='left', expand=True, fill="x")
        self.button.pack(padx=(5,0), side='left')
           
class LabeledCombo(ttk.Frame):
    def __init__(self, parent, labelText):
        super().__init__(master=parent)
        
        #self.configure(relief="groove")#for debugging
        
        self.label = ttk.Label(self, text=labelText)
        self.combo = ttk.Combobox(self)
        
        self.label.pack(padx=(0,5), side='left', expand=False, anchor='w')
        self.combo.pack(padx=(5,0), side='left', expand=True, fill="x")  

class NewExamGUI(tk.Toplevel):
    def __init__(self, parent):
        super().__init__()
        
        self.parent = parent
        
        self.title("Create new exam")
        self.geometry("500x200")
        
        self.courseNameCombo = LabeledCombo(self, "Select course")
        self.examWeekCombo = LabeledCombo(self, "Select exam week")
        self.examFilesURLLabel = LabeledEntry(self, "Enter exam files URL")
        
        self.createExamButton = ttk.Button(self, text="Create Exam", command=self.CreateNewCourse)
        self.courseNameCombo.combo['state'] = 'readonly'
        self.examWeekCombo.combo['state'] = 'readonly'
        
        
        
        self.courseNameCombo.pack(fill="both",padx=30, pady=(20,10))
        self.courseNameCombo.label.configure(width=18)
        
        self.examWeekCombo.pack(fill="both", padx=30, pady=10, anchor='w')
        self.examWeekCombo.label.configure(width=18)
        
        self.examFilesURLLabel.pack(fill="both", padx=30, pady=10)
        self.examFilesURLLabel.label.configure(width=18)
        
        self.createExamButton.pack(pady=10)
        
        
        courseData=toml.load("Courses.toml")
        self.courseNameCombo.combo['values'] = list(courseData.keys())
        self.courseNameCombo.combo.current(0)
        self.courseNameCombo.combo.bind('<<ComboboxSelected>>', self.OnCourseChange) 
        self.courseNameCombo.combo.bind('<Return>', self.CreateNewCourse) 
        self.OnCourseChange(None)
        self.examWeekCombo.combo.bind('<<ComboboxSelected>>', self.OnWeekChange) 
        self.examWeekCombo.combo.bind('<Return>', self.CreateNewCourse) 
    
    def OnCourseChange(self, event):
        
        courseData=toml.load("Courses.toml")
        course = self.courseNameCombo.combo.get()
        
        year = datetime.date.today().year
        currentWeek = datetime.date.today().isocalendar()[1]
        weeks = courseData[course]["reExamWeeks"]
        weeks.sort(reverse=True)
        
        weekTuples = [(year, week) if week >= currentWeek else (year+1, week) for week in weeks]
        sortedWeekTuples = sorted(weekTuples, key=lambda x: (x[1] - currentWeek) % 52)
        weekList = [f"{week} - {year}" for year, week in sortedWeekTuples]
        
        
        self.examWeekCombo.combo['values'] = weekList
        self.examWeekCombo.combo.current(0)
        
    def OnWeekChange(self, event):
        pass  
        
        
    def CreateNewCourse(self, event=None):
        course = self.courseNameCombo.combo.get()
        examWeekYear = self.examWeekCombo.combo.get()
        
        examWeekYear = examWeekYear.split(" - ")
        examYear = examWeekYear[1]
        examWeek = examWeekYear[0]

        
        courseData = toml.load("Courses.toml")
        courseName = courseData[course]["courseName"]
        courseCode = courseData[course]["courseCode"]
        canvasID = next( (canvasID for year,canvasID in courseData[course]["canvasIDs"] if year == int(examYear)), None )
        
        examID = f"{course}-{examWeek}-{examYear}"
        
        examFilesURL = self.examFilesURLLabel.entry.get()
        
        examData = toml.load("Exams.toml")
        
        examData[examID] = {}
        examData[examID]["courseName"] = courseName
        examData[examID]["examID"] = examID
        examData[examID]["courseCode"] = courseCode
        examData[examID]["canvasID"] = canvasID
        examData[examID]["year"] = examYear
        examData[examID]["week"] = examWeek
        examData[examID]["examFilesURL"] = examFilesURL

        
        with open('Exams.toml', 'w', encoding='utf-8') as file:
            toml.dump(examData, file)
        
        self.destroy()
        self.parent.UpdateWidgetsFromConfig()
        
        
        return
        

        self.config.add_section(courseID)
        self.config[courseID]["courseName"] = courseName
        self.config[courseID]["courseID"] = courseID
        self.config[courseID]["courseCode"] = courseCode
        self.config[courseID]["canvasCourseId"] = canvasCourseId
        self.config[courseID]["examfilesurl"] = examDesciptionURL
        self.config[courseID]["year"] = examYear
        self.config[courseID]["week"] = examWeek
        
        with open('config.ini', 'w', encoding='utf-8') as configfile:
            self.config.write(configfile)
        
        self.parent.courseCombo['values'] = self.config.sections() + ["New Course"]
        self.parent.courseCombo.current(len(self.config.sections())-1)
        
        

    
class ExamSchedulerGUI:
    def __init__(self):
        ## Window
        self.root = tk.Tk()
        width = 850
        height = 600
        self.root.geometry(f"{width}x{height}")
        self.root.minsize(width, height)
        self.root.title('Exam scheduler v0.3')

        ## Widgets
        self.InitializeWidgets()
        
        
        self.UpdateWidgetsFromConfig()
        self.root.mainloop()

    def InitializeWidgets(self):
        self.topFrame = ttk.Frame(self.root, padding=5)
        self.courseInfoFrame = ttk.Frame(self.root, padding=5, relief=tk.GROOVE)
        courseInfoLbl = ttk.Label(self.courseInfoFrame, text="Course info")
        
        self.selectedCourseString = tk.StringVar(value="course1")
        self.courseCombo = ttk.Combobox(self.topFrame, textvariable=self.selectedCourseString, width=30)
        self.courseCombo['values'] = ['course1', 'course2'] + ["New exam"]
        self.courseCombo.bind('<<ComboboxSelected>>', self.OnCourseComboChange) 
        self.courseCombo['state'] = 'readonly'
        
        self.courseNameLbl = ttk.Label(self.courseInfoFrame, text=f"Course name: ")
        self.courseCodeLbl = ttk.Label(self.courseInfoFrame, text=f"Course code: ")
        self.canvasIdLbl = ttk.Label(self.courseInfoFrame, text=f"Canvas ID: ", cursor="hand2")
        self.canvasIdLbl.bind("<ButtonRelease-1>", lambda e: webbrowser.open_new(f"https://ju.instructure.com/courses/"))
        
        self.studentSheetFrame = ttk.Frame(self.courseInfoFrame)
        self.studentListSheetLabel = ttk.Label(self.studentSheetFrame, text=f"Link: ", cursor="hand2")
        self.studentListSheetLabel.bind("<ButtonRelease-1>", lambda e: webbrowser.open_new("studentSheetUrl"))
        self.generateStudentListSheetBtn = ttk.Button(self.studentSheetFrame, text="Generate/update student list sheet", command=self.GenerateStudentListSheetButtonCallback)
        
        self.examInfoFrame = ttk.Frame(self.root, padding=5, relief=tk.GROOVE)
        examInfoLbl = ttk.Label(self.examInfoFrame, text="Exam info")
        self.examFilesUrlLabel = ttk.Label(self.examInfoFrame, text=f"Link to exam files: ", cursor="hand2")
        self.examFilesUrlLabel.bind("<ButtonRelease-1>", lambda e: webbrowser.open_new("examfilesurl"))
        
        self.composeExamEmailBtn = ttk.Button(self.examInfoFrame, text="Compose exam info email", command=self.ComposeExamEmailCallback)

        self.examFormFrame = ttk.Frame(self.examInfoFrame)
        self.generateCourseFormBtn = ttk.Button(self.examFormFrame, text="Generate course form", command=self.GenerateCourseFormButtonCallback)
        self.examFormLinkLabel = ttk.Label(self.examFormFrame, text=f"Link: ", cursor="hand2")
        self.examFormLinkLabel.bind("<ButtonRelease-1>", lambda e: webbrowser.open_new("examFormUrl"))
        
        self.sendFormBtn = ttk.Button(self.examInfoFrame, text="Send form", command=self.SendFormButtonCallback)
        self.formSentOnLabel = ttk.Label(self.examInfoFrame, text=f"Form sent on: ", background="")
        self.numberOfFormRecipientsLabel = ttk.Label(self.examInfoFrame, text=f"Number of recipients: ", background="")
        self.formClosingDateLabel = ttk.Label(self.examInfoFrame, text=f"Form closing date: ", background="")
        
        self.responseheetFrame = ttk.Frame(self.examInfoFrame)
        self.generateResponseSheetBtn = ttk.Button(self.responseheetFrame, text="Generate response sheet", command=self.GenerateResponseSheetButtonCallback)
        self.responseSheetLinkLabel = ttk.Label(self.responseheetFrame, text=f"Link: ", cursor="hand2")
        self.responseSheetLinkLabel.bind("<ButtonRelease-1>", lambda e: webbrowser.open_new("responseSheetUrl"))
        
        self.generateScheduleFrame = ttk.Frame(self.examInfoFrame)
        self.generateScheduleBtn = ttk.Button(self.generateScheduleFrame, text="Generate schedule sheet", command=self.GenerateScheduleButtonCallback)
        self.scheduleSheetLinkLabel = ttk.Label(self.generateScheduleFrame, text=f"Link: ", cursor="hand2")
        self.scheduleSheetLinkLabel.bind("<ButtonRelease-1>", lambda e: webbrowser.open_new("scheduleSheetUrl"))
        

        self.sendScheduleBtn = ttk.Button(self.examInfoFrame, text="Send schedule", command=self.SendScheduleButtonCallback)
        self.scheduleSentOnLabel = ttk.Label(self.examInfoFrame, text=f"Sent on: ", background="")
        self.numberOfScheduleRecipientsLabel = ttk.Label(self.examInfoFrame, text=f"Number of recipients: ", background="")
        
        self.removePassedStudentsFromListBtn = ttk.Button(self.examInfoFrame, text="Remove passed students", command=self.RemovePassedStudentsFromListButtonCallback)
        CreateToolTip(self.removePassedStudentsFromListBtn, "Remove passed students from the student list sheet.")
        
        self.statusText = tk.Text(self.root)
        
        ## Layout
        self.topFrame.pack(fill="both",padx=0, pady=5)
        self.courseCombo.pack(side="left", padx=5)
        self.courseInfoFrame.pack(fill="both",padx=10, pady=10)
        courseInfoLbl.place(x=0, y=-10)
        self.courseInfoFrame.columnconfigure((0,1,2,3,4), weight=1)
        self.courseNameLbl.grid(row=0, column=0, columnspan=3, pady=(15,5), sticky='w')
        self.courseCodeLbl.grid(row=0, column=3, pady=(15,5), sticky='w')
        self.canvasIdLbl.grid(row=0, column=4, pady=(15,5), sticky='w')
        
        self.studentSheetFrame.grid(row=1, column=0, columnspan=5, pady=(5,5), sticky='we')
        self.generateStudentListSheetBtn.pack(side="left")
        self.studentListSheetLabel.pack(side='left', expand=True, fill="x")
        
        self.examInfoFrame.pack(fill="both",padx=10, pady=10)
        examInfoLbl.place(x=0, y=-10)
        self.examInfoFrame.columnconfigure((0,1,2,3), weight=1)
        #self.courseInfoFrame.rowconfigure((0,1,2,3,4,5,6,7,8,9), weight=1)
        
        self.examFilesUrlLabel.grid(row=0, column=0, columnspan=4, pady=(15,5), sticky='we')

        self.composeExamEmailBtn.grid(row=10, column=0, columnspan=1, pady=(15,5), sticky='we')
        
        self.examFormFrame.grid(row=20, column=0, columnspan=4, pady=(5,5), sticky='we')
        self.generateCourseFormBtn.pack(side="left")
        #self.composeExamEmailBtn.pack(side="left")
        self.examFormLinkLabel.pack(side='left', expand=True, fill="x")
        
        self.sendFormBtn.grid(row=30, column=0, columnspan=1, pady=5, sticky='w')
        self.formSentOnLabel.grid(row=30, column=1, columnspan=1, pady=5, sticky='w')
        self.numberOfFormRecipientsLabel.grid(row=30, column=2, columnspan=1, pady=5, sticky='w')
        self.formClosingDateLabel.grid(row=30, column=3, columnspan=1, pady=5, sticky='w')
        
        self.responseheetFrame.grid(row=40, column=0, columnspan=4, pady=(5,5), sticky='we')
        self.generateResponseSheetBtn.pack(side="left")
        self.responseSheetLinkLabel.pack(side='left', expand=True, fill="x")
        
        self.generateScheduleFrame.grid(row=50, column=0, columnspan=4, pady=(5,5), sticky='we')
        self.generateScheduleBtn.pack(side="left")
        self.scheduleSheetLinkLabel.pack(side='left', expand=True, fill="x")

        self.sendScheduleBtn.grid(row=60, column=0, columnspan=1, pady=5, sticky='w')
        self.scheduleSentOnLabel.grid(row=60, column=1, columnspan=1, pady=5, sticky='w')
        self.numberOfScheduleRecipientsLabel.grid(row=60, column=2, columnspan=1, pady=5, sticky='w')
        
        self.removePassedStudentsFromListBtn.grid(row=70, column=0, columnspan=1, pady=5, sticky='w')
        
        self.statusText.pack(pady=10,padx=10,fill='both')
        
        
        self.examData = toml.load("Exams.toml")
        examData = self.examData
        courseList = list(examData.keys())
        self.courseCombo['values'] = courseList
        self.courseCombo.current(len(courseList)-1)
        self.courseCombo['values'] = courseList + ["New exam"]
    
    def OnCourseComboChange(self, event):
        if self.selectedCourseString.get() == "New exam":
            self.OpenNewExamWindow()
        else:
            self.UpdateWidgetsFromConfig()
    
    def UpdateWidgetsFromConfig(self):
        print("Updating all widgets")
        
        examData = toml.load("Exams.toml")
        courseData = toml.load("Courses.toml")
        
        courseList = list(examData.keys())
        self.courseCombo['values'] = courseList + ["New exam"]
        
        
        selectedCourseString = self.selectedCourseString.get()
        if selectedCourseString == "New exam":
            selectedCourseString = courseList[-1]
            self.courseCombo.current(len(courseList)-1)
            
        selectedCourse = selectedCourseString.split('-')[0]
        examFilesURL = TryReadExamConfig(self, examData, selectedCourseString, "examFilesURL")
        canvasID = TryReadExamConfig(self, examData, selectedCourseString, "canvasID")
        courseCode = TryReadExamConfig(self, examData, selectedCourseString, "courseCode")
        courseName = TryReadExamConfig(self, examData, selectedCourseString, "courseName")

        self.courseNameLbl.configure(text=f"Course name: {courseName}")
        self.courseCodeLbl.configure(text=f"Course code: {courseCode}")
        self.canvasIdLbl.configure(text=f"Canvas ID: {canvasID}")
        self.canvasIdLbl.bind("<ButtonRelease-1>", lambda e: webbrowser.open_new(f"https://ju.instructure.com/courses/{canvasID}"))
        
        self.examFilesUrlLabel.configure(text=f"Link to exam files: {examFilesURL}")
        if not examFilesURL == "":
            self.examFilesUrlLabel.bind("<ButtonRelease-1>", lambda e: webbrowser.open_new(examFilesURL))
        else:
            self.examFilesUrlLabel.bind("<ButtonRelease-1>", lambda e:None)

        if "studentSheetURL" in courseData[selectedCourse]:
            studentSheetURL = courseData[selectedCourse]["studentSheetURL"]  
            self.studentListSheetLabel.bind("<ButtonRelease-1>", lambda e: webbrowser.open_new(studentSheetURL))  
        else:
            studentSheetURL = ""
            self.studentListSheetLabel.bind("<ButtonRelease-1>", lambda e: None)  
        self.studentListSheetLabel.configure(text=f'Link: {studentSheetURL}')
          

        if "formUri" in examData[selectedCourseString]:
            formUri = examData[selectedCourseString]["formUri"]
            self.examFormLinkLabel.bind("<ButtonRelease-1>", lambda e: webbrowser.open_new(formUri))
        else:
            formUri = ""
            self.examFormLinkLabel.bind("<ButtonRelease-1>", lambda e: None)
        self.examFormLinkLabel.configure(text=f'Link: {formUri}')    
            
        if "responseSheetUri" in examData[selectedCourseString]:
            responseSheetUri = examData[selectedCourseString]["responseSheetUri"]
            self.responseSheetLinkLabel.bind("<ButtonRelease-1>", lambda e: webbrowser.open_new(responseSheetUri))    
        else:
            responseSheetUri=""
            self.responseSheetLinkLabel.bind("<ButtonRelease-1>", lambda e: None)    
        self.responseSheetLinkLabel.configure(text=f'Link: {responseSheetUri}')
        
        if "formSentDate" in examData[selectedCourseString]:
            formSentDate = examData[selectedCourseString]["formSentDate"]
        else:
            formSentDate = ""
        self.formSentOnLabel.configure(text=f'Form sent on: {formSentDate}')
        
        if "totalNumberOfStudents" in examData[selectedCourseString]:
            totalNumberOfStudents = examData[selectedCourseString]["totalNumberOfStudents"]
        else:
            totalNumberOfStudents = ""
        self.numberOfFormRecipientsLabel.configure(text=f'Number of recipients: {totalNumberOfStudents}')
        
        if "formClosingDate" in examData[selectedCourseString]:
            formClosingDate = examData[selectedCourseString]["formClosingDate"]
        else:
            formClosingDate = ""
        self.formClosingDateLabel.configure(text=f'Form closing date: {formClosingDate}')
        
        if "examSheetUri" in examData[selectedCourseString]:
            examSheetUri = examData[selectedCourseString]["examSheetUri"]
        else:
            examSheetUri = ""
        self.scheduleSheetLinkLabel.configure(text=f'Link: {examSheetUri}')
        self.scheduleSheetLinkLabel.bind("<ButtonRelease-1>", lambda e: webbrowser.open_new(examSheetUri))
        
        
        
        if "scheduleSentDate" in examData[selectedCourseString]:
            scheduleSentDate = examData[selectedCourseString]["scheduleSentDate"]
        else:
            scheduleSentDate = ""
        self.scheduleSentOnLabel.configure(text=f'Sent on: {scheduleSentDate}')
        
        if "numberOfScheduleRecipients" in examData[selectedCourseString]:
            numberOfScheduleRecipients = examData[selectedCourseString]["numberOfScheduleRecipients"]
        else:
            numberOfScheduleRecipients = ""
        self.numberOfScheduleRecipientsLabel.configure(text=f'Number of recipients: {numberOfScheduleRecipients}')
        
        pass
    
    def GenerateStudentListSheetButtonCallback(self):
        courseData= toml.load("Courses.toml")
        selectedCourseString = self.selectedCourseString.get()
        course = selectedCourseString.split('-')[0]
        print(course)
        GenerateStudentsSheet(course, courseData)
        self.UpdateWidgetsFromConfig()
        print("Done!")

    def GenerateCourseFormButtonCallback(self):
        print("Generating exam form...")
        selectedCourseString = self.selectedCourseString.get()
        CreateExamForm(selectedCourseString)
        self.UpdateWidgetsFromConfig()
        pass

    def ComposeExamEmailCallback(self):
        print("Composing email and opening Outlook...")
        examID = self.selectedCourseString.get()
        ComposeExamEmail(examID)
        self.UpdateWidgetsFromConfig()
    
    def GenerateResponseSheetButtonCallback(self):
        print("Generating exam sheet...")
        selectedCourseString = self.selectedCourseString.get()
        GenerateResponseSheet(selectedCourseString)
        self.UpdateWidgetsFromConfig()
        pass

    def SendFormButtonCallback(self):
        print("SendFormBtnButtonCallback")
        examID = self.selectedCourseString.get()
        SendFormToAllStudent(examID)
        self.UpdateWidgetsFromConfig()
        pass
    
    def GenerateScheduleButtonCallback(self):
        print("Generating schedule")
        examID = self.selectedCourseString.get()
        GenerateSchedule(examID)
        self.UpdateWidgetsFromConfig()
    
    def SendScheduleButtonCallback(self):
        print("Sending schedule...")
        examID = self.selectedCourseString.get()
        SendExamSchedule(examID)
        self.UpdateWidgetsFromConfig()
        pass
    
    def RemovePassedStudentsFromListButtonCallback(self):
        print("RemovePassedStudentsFromListButtonCallback")
        pass

    def OpenNewExamWindow(self):
        global newExamWindow
        newExamWindow = NewExamGUI(self)

    def CloseNewCourse(self):
        newExamWindow.destroy()


if __name__ == "__main__":
    ExamSchedulerGUI()