# Installaton

## 1. Conda installation

1. Here, we shall use conda for virtual environments, [install anaconda](https://www.anaconda.com/download), making sure to add `conda` to the path during install, verify that the command `conda` can be executed from powershell.

2. Run  `conda update conda` and then `conda init` from powershell. This activates the base environment and gives acces to python and pip.

3. Restart powershell and you are set to use environments with conda!

## 2. Installing requirements

To install required packages run

```
pip install -r requirements.txt
```


## In case of error messages at powerscript start up: 

````
profile.ps1 cannot be loaded because running scripts is disabled on this system.
````

The issue is that PowerShell by default restricts running scripts unless we change the [execution policies](https://learn.microsoft.com/en-us/powershell/module/microsoft.powershell.core/about/about_execution_policies?view=powershell-7.1#bypass).

we can change the execution policies by adding `-ExecutionPolicy Bypass` args. in vscode you will have to create a shell profile with these args by adding below in settings.json (ctrl + shift + p and type "settings.json" choose `Open User Settings`)

Add the following

```
"terminal.integrated.profiles.windows": {
  "PowerShell": {
    "source": "PowerShell",
    "icon": "terminal-powershell",
    "args": ["-ExecutionPolicy", "Bypass"]
  }
},
"terminal.integrated.defaultProfile.windows": "PowerShell",
```

Restart VS Code and the Terminal.

[Source](https://stackoverflow.com/questions/56199111/visual-studio-code-cmd-error-cannot-be-loaded-because-running-scripts-is-disabl/67420296#67420296)

# Environments

## Creating an environment from the `environment.yml` file

See [starting fresh](#Starting-fresh) if you want to generate a new environment.

To use an existing environment, use the terminal for the following steps:

1. Create the environment from the `environment.yml` file:

```
conda env create -f environment.yml
```

The first line of the `yml` file sets the new environment's name. For details see [Creating an environment file manually](#Sharing-environments).

2. Activate the new environment: `conda activate AutoFormingSim`
3. Verify that the new environment was installed correctly: `conda env list` and `conda list`



## Sharing environments 

`conda env export > environment.yml`

This file handles both the environment's pip packages and conda packages.



## Starting fresh

To create a fresh and empty environment run 

`conda create -n AutoFormingSim python=3.6`

`conda activate AutoFormingSim`

The expected output is something like this:

```
(c:\Users\mirza\Dropbox\Work\Digitalization\AutoFormingSim\env) PS C:\Users\mirza\Dropbox\Work\Digitalization\AutoFormingSim>
```

To shorten the environment prefix run `conda config --set env_prompt '({name})'`, this will yield:

```
(env)PS C:\Users\mirza\Dropbox\Work\Digitalization\AutoFormingSim> 
```

`conda deactivate`

`conda create --prefix ./env36 python=3.6`

## Using `create-env.txt`

`conda create -n AutoFormingSim --file create-env.txt`

Note, this does not inclde packages, use the `environment.yml` to copy both environments and packages.

## Create environment file

`conda list --explicit > create-env.txt`

## Usage

List all installed packages in *active* environment

`conda list`

# Generating requirements.txt

Install pipreqs
```
pip install pipreqs
```

Use `pipreqs`

```
pipreqs --print --use-local
```

ommit the `--print` to generate the requirements.txt file.