
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build
from googleapiclient.errors import HttpError
import toml #type: ignore stupid error message
from canvasapi import Canvas

import win32com.client as win32
import os.path
import datetime
import sys
import tkinter as tk
from tkinter import ttk
import webbrowser
from tkinter import messagebox
import calendar





def Int2ColumnLetter(column_int, start_index = 1):
    # start_index = 1   #  it can start either at 0 or at 1
    letter = ''
    while column_int > 25 + start_index:   
        letter += chr(65 + int((column_int-start_index)/26) - 1)
        column_int = column_int - (int((column_int-start_index)/26))*26
    letter += chr(65 - start_index + (int(column_int)))
    return letter


def SendEmail(to, subject, message):
    outlook = win32.Dispatch('outlook.application')
    mail = outlook.CreateItem(0)
    mail.To = to
    mail.Subject = subject
    mail.Body = message
    #mail.HTMLBody = '<h2>HTML Message body</h2>' #this field is optional

    # To attach a file to the email (optional):
    #attachment  = "Path to the attachment"
    #mail.Attachments.Add(attachment)

    mail.Send()

def AuthenticateAgainstGoogle():
    # If modifying these scopes, delete the file token.json.
    SCOPES = ['https://www.googleapis.com/auth/spreadsheets', 
            'https://www.googleapis.com/auth/drive.readonly',
            'https://www.googleapis.com/auth/forms.body']
    def login():
        flow = InstalledAppFlow.from_client_secrets_file('credentials.json', SCOPES)
        creds = flow.run_local_server(port=0)
        # Save the credentials for the next run
        with open('token.json', 'w') as token:
            token.write(creds.to_json())
        return creds
            
    creds = None
        # The file token.json stores the user's access and refresh tokens, and is
        # created automatically when the authorization flow completes for the first
        # time.
    if os.path.exists('token.json'):
        creds = Credentials.from_authorized_user_file('token.json', SCOPES)
        # If there are no (valid) credentials available, let the user log in.
    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            try:
                creds.refresh(Request())
            except:
                creds = login()
        else:
            creds = login()
    return creds     

def CreateExamForm(examID):
    # selectedCourseString = "MekHallf2-33-2023"
    
    courseData = toml.load("Courses.toml")
    examData = toml.load("Exams.toml")
    
    if 'formUri' in examData[examID]:
        formUri = examData[examID]['formUri']
        responderUri = examData[examID]['responderUri']
        print("Form already exists!")
        print(f"Form edit link: {formUri}") 
        print(f"Form responder link: {responderUri}") 
        return
    
    course = examID.split('-')[0]
    week = examID.split('-')[1]
    year = examID.split('-')[2]
    
    courseName = courseData[course]["courseName"]
    
    
    title = f"{courseName} exam week {week} {year}"

    # Check if form already exists
    driveService = build('drive', 'v3', credentials=AuthenticateAgainstGoogle())
    formsService = build('forms', 'v1', credentials=AuthenticateAgainstGoogle())

    print(f'Checking if form title: "{title}" already exists on the drive...')

    query = f'trashed=false and mimeType = "application/vnd.google-apps.form" and name="{title}"'
    results = driveService.files().list(q=query, fields="nextPageToken, files(id, name)").execute()
    filesResult = results['files']
    if(len(filesResult) != 0):
        formID = filesResult[0]["id"]
        print(f'Form with title "{title}" already exists! Returning its ID: {formID}')
        result = formsService.forms().get(formId=formID).execute()
        responderUri = result.get("responderUri")
        formUri = f"https://docs.google.com/forms/d/{formID}/edit"
        examData[examID]["formID"] = formID
        examData[examID]["responderUri"] = responderUri
        examData[examID]["formUri"] = formUri
        WriteExamData(examData)
        print(f"Form URI: {formUri}")
        print(f"Form reponder URI: {responderUri}")
        return
    else:
        print("Form does not exist, creating a new form...")

    # Generate new Google form
    newForm = {
        "info": {
            "title": title,
            "documentTitle": title
        }
    }
    

    #currentYear = datetime.date.today().year
    currentYear = int(year)
    years = [{'value':str(year)} for year,canvasID in courseData[course]["canvasIDs"]]

    examFilesURL = examData[examID]["examFilesURL"]
    deadlineDate = datetime.date.today() + datetime.timedelta(weeks=2)
    formClosingDate = deadlineDate.strftime('%Y-%m-%d')
    print(f"formClosingDate: {formClosingDate}")

    update = {
        "requests": [{
            "updateFormInfo": {
                "info": {
                    "title": title,
                    "description": f"Form registration for the course {courseName} for week {week} {year}.\nGet the exam files here: {examFilesURL}\nUse the 'Other info' to let the examiner know about any critical scheduling conflicts during the exam week.\nThe registration form closes on {formClosingDate} after which you will receive an email with your exam day, place and time."
                },
                "updateMask": "*"
            }            
        },{
            "createItem":{
                "item": {
                    "title": "Name",
                    "questionItem": {
                        "question": {
                            "required": True,
                            "textQuestion": {
                                "paragraph": False
                            }
                        }
                    }
                },
                "location": {
                    "index": 0
                }
            }
        },{
            "createItem":{
                "item": {
                    "title": "Last name",
                    "questionItem": {
                        "question": {
                            "required": True,
                            "textQuestion": {
                                "paragraph": False
                            }
                        }
                    }
                },
                "location": {
                    "index": 1
                }
            }
        },{
            "createItem":{
                "item": {
                    "title": "Email (student email)",
                    "questionItem": {
                        "question": {
                            "required": True,
                            "textQuestion": {
                                "paragraph": False
                            }
                        }
                    }
                },
                "location": {
                    "index": 2
                }
            }
        },{
           "createItem": {
                "item": {
                    "title": "When did you have this course?",
                    "questionItem": {
                        "question": {
                            "required": True,
                            "choiceQuestion": {
                                "type": "DROP_DOWN",
                                "options": years,
                                "shuffle": False
                            }
                        }
                    },
                },
                "location": {
                    "index": 3
                }
            } 
        },{
            "createItem":{
                "item": {
                    "title": "Other info to the examiner",
                    "questionItem": {
                        "question": {
                            "required": False,
                            "textQuestion": {
                                "paragraph": False
                            }
                        }
                    }
                },
                "location": {
                    "index": 4
                }
            }
        }]
    }

    
    print("Generating Google form...")
    result = formsService.forms().create(body=newForm).execute()
    formID=result["formId"]
    
    formsService.forms().batchUpdate(formId=formID, body=update).execute()


    # Prints the result to show the question has been added
    get_result = formsService.forms().get(formId=formID).execute()
    responderUri = get_result.get("responderUri")
    
    formUri = f"https://docs.google.com/forms/d/{formID}/edit"
    examData[examID]["formID"] = formID
    examData[examID]["responderUri"] = responderUri
    examData[examID]["formUri"] = formUri
    examData[examID]["formClosingDate"] = formClosingDate
    WriteExamData(examData)
    print(f"Form URI: {formUri}")
    print(f"Form reponder URI: {responderUri}")
    
    return

def ComposeEmailToStudent(name, lastname, studyYear, year, week, courseName, formClosingDate, examFilesURL, responderUri, courseCode):
    emailMsg = \
        f'Hello {name.strip()} {lastname.strip()}!\n\n' \
        f'According to my notes you have not yet passed the course {courseName} - {courseCode}.\n\n' \
        f'You either have labs/ work package left to present or you did not pass the exam.\n' \
        f'If you only have labs/ work packages left to present, please contact us during the next period to arrange for a presentation during work hours.\n\n' \
        f'If you have the examination left to do, then the next examination for the course {courseName} is in week {week} {year}.\n' \
        f'All exam files and instructions can be found here: {examFilesURL}\n' \
        f'Examination will be done orally at JTH, more info on the exact time and place will come in a later email if you register for the exam.\n' \
        f'You can register to the exam at this link: {responderUri}\n\n' \
        f'At the examination you will have 10 minutes to present your exam problem, solution and analysis using an individual presentation as aid. ' \
        f'After that 10 additional minutes are allotted to questions by the examiners regarding your solution, problem modeling, setup, analysis, and conclusions.\n\n' \
        f'The deadline for submitting the form is {formClosingDate}\n\n\n' \
        f'If you have received this email in error, please notify Mirza.\n\n' \
        f'Best regards,\n' \
        f'Mirza Cenanovic\n\n\n' \
        f'This message was autogenerated.'
    return emailMsg

def SendFormToAllStudent(examID):
    
    course = examID.split('-')[0]
    week = examID.split('-')[1]
    year = examID.split('-')[2]
    courseData = toml.load("Courses.toml")
    examData = toml.load("Exams.toml")
    courseName = courseData[course]["courseName"]
    courseCode = courseData[course]["courseCode"]
    examFilesURL = examData[examID]["examFilesURL"]
    responderUri = examData[examID]["responderUri"]
    formClosingDate = examData[examID]["formClosingDate"]

    studentSheetID = courseData[course]['studentSheetID']
    sheetsService = build('sheets', 'v4', credentials=AuthenticateAgainstGoogle())
    studentSheet = sheetsService.spreadsheets().values().get(spreadsheetId=studentSheetID, range='Sheet1').execute()
    studentSheetValues = studentSheet['values']
    
    nStudents = len(studentSheetValues)-1
    examData[examID]["totalNumberOfStudents"] = nStudents
    WriteExamData(examData)
    
    webbrowser.open_new(courseData[course]['studentSheetURL'])
    
    if not messagebox.askyesno("Send form to students?", f"Do you want to send emails with the registration form to all {nStudents} students?"):
        print("Nothing is sent to the students.")
        return
    
    formSentDate = datetime.date.today().strftime('%Y-%m-%d')
    examData[examID]["formSentDate"] = formSentDate
    WriteExamData(examData)

    emailSubject = f'{courseName} - upcoming exam registration form'
    for i, row in enumerate(studentSheetValues):
        if len(row)==0:
            continue
        if row[0]=="Name":
            #print('Name      |  Last name       |  Email                 |  Year  | Hinder | Email Sent')
            #print('------------------------------------------------------------------------------------')
            continue
        name = row[0].strip()
        lastname = row[1].strip()
        email = row[2] 
        studyYear = row[3]  
        
        hinder = "" 
        if len(row) > 4:
            hinder = row[4]
        
        emailPreviouslySent = "" 
        if len(row) > 5:
            emailPreviouslySent = row[5]  
        
        #print(f'{name}   | {lastname}  | {email} | {studyYear} | {hinder} | {emailPreviouslySent}')
        
        if not hinder.lower() == "no":
            print(f"Hinder is not set to 'no', NOT sending email to {name} {lastname}!")
            continue
        if not emailPreviouslySent == "":
            print(f"Email was previously sent to {name} {lastname}!")
            continue
        
        print(f"Sending email to {name} {lastname}...")
        emailMsg = ComposeEmailToStudent(name, lastname, studyYear, year, week, courseName, formClosingDate, examFilesURL, responderUri, courseCode)
        print(emailMsg)  
        SendEmail(email, emailSubject, emailMsg)
        sheetsService.spreadsheets().values().update( spreadsheetId=studentSheetID, range=f"Sheet1!F{i+1}:F{i+1}" , valueInputOption="USER_ENTERED", body={'values': [[formSentDate]] }).execute()
        sheetsService.spreadsheets().values().update( spreadsheetId=studentSheetID, range=f"Sheet1!E{i+1}:E{i+1}" , valueInputOption="USER_ENTERED", body={'values': [['']] }).execute()
    
    
    
    

def CreateEmail(to, subject, message):
    outlook = win32.Dispatch('outlook.application')
    mail = outlook.CreateItem(0)
    mail.To = to
    mail.Subject = subject
    mail.Body = message
    mail.Save()
    mail.Display(False)

def GenerateResponseSheet(examID):
    
    courseData = toml.load("Courses.toml")
    examData = toml.load("Exams.toml")
    
    if 'responseSheetID' in examData[examID]:
        responseSheetID = examData[examID]['responseSheetID']
        print("Sheet already exists!")
        responseSheetUri = f"https://docs.google.com/spreadsheets/d/{responseSheetID}/edit"
        print(f"Response sheet link: {responseSheetUri}")   
        return
    
    # Create the exam sheet
    formsService = build('forms', 'v1', credentials=AuthenticateAgainstGoogle())
    formID = examData[examID]['formID']
    courseForm = formsService.forms().get(formId=formID).execute()
    if 'linkedSheetId' in courseForm:
        responseSheetID = courseForm['linkedSheetId']
        responseSheetUri = f"https://docs.google.com/spreadsheets/d/{responseSheetID}/edit"
        examData[examID]['responseSheetID'] = responseSheetID
        examData[examID]['responseSheetUri'] = responseSheetUri
        WriteExamData(examData)
        print(f"Response sheet link: {responseSheetUri}")  
        return
    
    webbrowser.open_new(f"https://docs.google.com/forms/d/{formID}/edit#responses")
        
    print(f'Responses not generated in a google sheet, visit https://docs.google.com/forms/d/{formID}/edit#responses and click "Link to Sheets"')
    if not messagebox.askyesno("Response sheet exists?", f"Is the sheet created now?"):
        print("User aborted. Exam sheet not created.")
        return
    
    # Update courseForm to get the newest data
    courseForm = formsService.forms().get(formId=formID).execute()  
    if 'linkedSheetId' not in courseForm:
        print("The response sheet is still not generated.")
        return
    
    responseSheetID = courseForm['linkedSheetId']
    responseSheetUri = f"https://docs.google.com/spreadsheets/d/{responseSheetID}/edit"
    examData[examID]['responseSheetID'] = responseSheetID
    examData[examID]['responseSheetUri'] = responseSheetUri
    WriteExamData(examData)
    print(f"Response sheet link: {responseSheetUri}")    
    return

def GenerateSchedule(examID):
    
    course = examID.split('-')[0]
    week = examID.split('-')[1]
    year = examID.split('-')[2]
    
    courseData = toml.load("Courses.toml")
    examData = toml.load("Exams.toml")
    currentExamIDs = [examID for examID in examData if int(examData[examID]["week"]) == int(week) ]
    print(currentExamIDs)
    
    # Read or create exam sheet
    examRoom = "TBD"
    examSheetName = f'Exam week {week} {year}'
    driveService = build('drive', 'v3', credentials=AuthenticateAgainstGoogle())
    sheetsService = build('sheets', 'v4', credentials=AuthenticateAgainstGoogle())
    query = f'trashed=false and mimeType = "application/vnd.google-apps.spreadsheet" and name="{examSheetName}"'
    sheetFile = driveService.files().list(q=query, fields="nextPageToken, files(id, name)").execute()
    if sheetFile['files']:
        print(f"Google sheet file '{examSheetName}' already exists!")
        spreadSheetID = sheetFile["files"][0]["id"]
    else:
        print(f"Generating new exam sheet: '{examSheetName}' ")
        spreadSheet = {
                        'properties': {
                            'title': examSheetName
                        }
                    }
        spreadSheet = sheetsService.spreadsheets().create(body=spreadSheet, fields='spreadsheetId').execute()
        spreadSheetID = spreadSheet.get('spreadsheetId')
        
    
    examSheetUri = f"https://docs.google.com/spreadsheets/d/{spreadSheetID}/edit#gid=0"
    print(f"Exam sheet URI: {examSheetUri}")
    webbrowser.open_new(examSheetUri)
    

  

    # Looking for unique students from the response sheets...
    print("Looking for unique students from the response sheets...")
    (scheduleStudentsData, examRoom) = GetScheduledStudents(sheetsService, spreadSheetID)

    newStudents = GetUniqueRespondingStudents(currentExamIDs, examData, sheetsService, scheduleStudentsData)
    
    print(f"{len(newStudents)} new students found in the response sheets.")
    if len(newStudents) == 0:
        print("No new students found in the response sheets. Nothing is added to the exam schedule!")
        return

    
    
    if not messagebox.askyesno("Update exam schedule?", f"Do you want to update the existing exam schedule with {len(newStudents)} new students?"):
        print("Exam schedule not updated!")
        return
    
    
    startDay = 1 # Monday
    examDate = GetDateOfWeekNumber(year, week, targetDay=startDay)
    dayList = ['Monday','Tuesday','Wednesday','Thursday','Friday','Saturday','Sunday']
    day = dayList[examDate.isocalendar().weekday-1]
    examDayString = f"{day} {calendar.month_name[examDate.month]} {examDate.day}"

    sheetData = sheetsService.spreadsheets().values().get(spreadsheetId=spreadSheetID, range='Sheet1').execute()
    if "values" in sheetData:
        values = sheetData['values']
        (schedule, startTime) = RecreateSchedule(values, year, week, dayList)
    else:
        values = [['Day',examDayString],
              ['Start Time', 'End Time', 'Name', 'Last Name', 'Email', 'Study Year', 'Comment', 'Course', 'Exam ID', 'Exam Room', ' Email Sent', 'LADOK', 'Grade', 'Grade Comment']]
        startTime = datetime.datetime(2023, 5, 11, 8, 15)  # start time
    
    endTime = datetime.datetime(2023, 5, 11, 17, 0)  # end time    

    #sheetsService.spreadsheets().values().update( spreadsheetId=spreadSheetID, range=f"Sheet1!A1:N2", valueInputOption="USER_ENTERED", body={'values': values }).execute()
  
    # Loop over all response data
    for examID in currentExamIDs:
        course = examID.split('-')[0]
        examData[examID]["examSheetUri"] = examSheetUri
        examData[examID]["examSheetID"] = spreadSheetID
        WriteExamData(examData)

    
    for c, student in enumerate(newStudents):
        name = student[0]
        lastName = student[1]
        email = student[2]
        studyYear = student[3]
        comment = student[4]
        studentExamID = student[6]

        if c == 0:
            prevExamID = values[-1][8]

        # Schedule
        if startTime > endTime:
            startTime = datetime.datetime(2023, 5, 11, 8, 15)
            schedule = [] 
            startDay += 1
            examDate = GetDateOfWeekNumber(year, week, targetDay=startDay)
            day = dayList[examDate.isocalendar().weekday-1]
            examDayString = f"{day} {calendar.month_name[examDate.month]} {examDate.day}"
            values.append([]) 
            values.append([]) 
            values.append(['Day',examDayString])
            continue
        
        if(startTime.hour == 12):
            startTime = datetime.datetime(2023, 5, 11, 13, 15)
        startTimeString = startTime.strftime('%H:%M')
        startTime += datetime.timedelta(minutes=20)  # increment start time by 20 minutes
        
            
        endTimeString = startTime.strftime('%H:%M')
        schedule.append([startTimeString, endTimeString])  # add start time to the list
        
        if(len(schedule) % 4 == 0):
            startTime += datetime.timedelta(minutes=20)  # add 20 minute break
        else:
            startTime += datetime.timedelta(minutes=5)  # add 20 minute setup time
        
        if not prevExamID == studentExamID:
            values.append([]) 
        values.append([startTimeString, endTimeString, name, lastName, email, studyYear, comment, course, studentExamID, examRoom, "", "", "", ""])  
        prevExamID = studentExamID
        
        
           
    
    nRows = len(values)
    range = f"Sheet1!A1:N{nRows}"
    sheetsService.spreadsheets().values().update( spreadsheetId=spreadSheetID, range=range, valueInputOption="USER_ENTERED", body={'values': values }).execute()
    print("Schedule has been generated.")
    return

def RecreateSchedule(values, year, week, dayList):
    startTime = datetime.datetime(2023, 5, 11, 8, 15)  # start time
    endTime = datetime.datetime(2023, 5, 11, 17, 0)  # end time
    # Grab schedule
    schedule = []  # empty list to store start and end times
    for row in values:
        if len(row) < 1: continue
        if row[0] == "Day": continue
        if row[0] == "Start Time": continue

        startTimeString = startTime.strftime('%H:%M')
        startTime += datetime.timedelta(minutes=20)  # increment start time by 20 minutes  
        endTimeString = startTime.strftime('%H:%M')
        schedule.append([startTimeString, endTimeString])  # add start time to the list

        if(len(schedule) % 4 == 0):
            startTime += datetime.timedelta(minutes=20)  # add 20 minute break
        else:
            startTime += datetime.timedelta(minutes=5)  # add 20 minute setup time

        if(startTime.hour == 12):
            startTime = datetime.datetime(2023, 5, 11, 13, 15)
        
        if startTime > endTime:
            startTime = datetime.datetime(2023, 5, 11, 8, 15)
            schedule = [] 
            startDay += 1
            examDate = GetDateOfWeekNumber(year, week, targetDay=startDay)
            day = dayList[examDate.isocalendar().weekday-1]
            examDayString = f"{day} {calendar.month_name[examDate.month]} {examDate.day}"
            continue
    return (schedule, startTime)


def GetUniqueRespondingStudents(currentExamIDs, examData, sheetsService, scheduleStudentsData):

    responseStudentsData = []
    for examID in currentExamIDs:
        course = examID.split('-')[0]
        responseSheetID = examData[examID]["responseSheetID"]
        sheetData = sheetsService.spreadsheets().values().get(spreadsheetId=responseSheetID, range='Form Responses 1').execute()
        sheetData = sheetData['values']

        for row in sheetData:
            if row[0] == "Timestamp": continue
            name = row[1]
            lastName = row[2]
            email = row[3]
            studyYear = row[4]
            comment = ""
            if len(row) == 6:
                comment = row[5]

            responseStudentsData.append([ name, lastName, email, studyYear, comment, course, examID ])
            #print(f"{name} {lastName}, {email}, {studyYear}, {examID}")

    newStudents = []
    for respondingStudent in responseStudentsData:
        email = respondingStudent[2]
        examID = respondingStudent[6]
        if not next((student for student in scheduleStudentsData if(email in student and examID in student)), []):
            newStudents.append(respondingStudent)
            print(f"{respondingStudent[0]} {respondingStudent[1]}, {respondingStudent[2]}, {respondingStudent[3]}, {respondingStudent[5]}")

    return newStudents

        

def GetScheduledStudents(sheetsService, spreadSheetID):
    scheduleStudentsData = []  
    examRoom = "TBD"
    sheetData = sheetsService.spreadsheets().values().get(spreadsheetId=spreadSheetID, range='Sheet1').execute()
    if not "values" in sheetData:
        return (scheduleStudentsData, examRoom)
    
    sheetData = sheetData['values']
    if len(sheetData) > 3:
        examRoom = sheetData[2][9]
      
    for row in sheetData:
        if len(row) < 9: continue
        if row[0] == "Day": continue
        if row[0] == "Start Time": continue
        
        scheduleStudentsData.append([ row[2], row[3], row[4], row[5], row[6], row[7], row[8] ])

    return (scheduleStudentsData, examRoom)

# https://stackoverflow.com/questions/3221956/how-do-i-display-tooltips-in-tkinter
class CreateToolTip(object):
    """
    create a tooltip for a given widget
    """
    def __init__(self, widget, text='widget info'):
        self.waittime = 500     #miliseconds
        self.wraplength = 180   #pixels
        self.widget = widget
        self.text = text
        self.widget.bind("<Enter>", self.enter)
        self.widget.bind("<Leave>", self.leave)
        self.widget.bind("<ButtonPress>", self.leave)
        self.id = None
        self.tw = None

    def enter(self, event=None):
        self.schedule()

    def leave(self, event=None):
        self.unschedule()
        self.hidetip()

    def schedule(self):
        self.unschedule()
        self.id = self.widget.after(self.waittime, self.showtip)

    def unschedule(self):
        id = self.id
        self.id = None
        if id:
            self.widget.after_cancel(id)

    def showtip(self, event=None):
        x = y = 0
        x, y, cx, cy = self.widget.bbox("insert")
        x += self.widget.winfo_rootx() + 25
        y += self.widget.winfo_rooty() + 20
        # creates a toplevel window
        self.tw = tk.Toplevel(self.widget)
        # Leaves only the label and removes the app window
        self.tw.wm_overrideredirect(True)
        self.tw.wm_geometry("+%d+%d" % (x, y))
        label = tk.Label(self.tw, text=self.text, justify='left',
                       background="#ffffff", relief='solid', borderwidth=1,
                       wraplength = self.wraplength)
        label.pack(ipadx=1)

    def hidetip(self):
        tw = self.tw
        self.tw= None
        if tw:
            tw.destroy()


def ReadCanvasToken():
    with open('privateCanvasToken', 'r') as f:
        lines = f.read()
    return lines

def GenerateStudentsSheet(courseName, courseData, updateStudentsSheet=False):
    canvas = Canvas("https://ju.instructure.com/", ReadCanvasToken())
    canvasCourseIDs = [canvasID for year,canvasID in courseData[courseName]["canvasIDs"]]


    # Read or create student sheet
    sheetName = f'{courseName} Students'
    driveService = build('drive', 'v3', credentials=AuthenticateAgainstGoogle())
    sheetsService = build('sheets', 'v4', credentials=AuthenticateAgainstGoogle())
    query = f'trashed=false and mimeType = "application/vnd.google-apps.spreadsheet" and name="{sheetName}"'
    sheetFile = driveService.files().list(q=query, fields="nextPageToken, files(id, name)").execute()
    if sheetFile['files']:
        print(f"google sheet file '{sheetName}' already exists!")
        spreadSheetID = sheetFile["files"][0]["id"]
    else:
        print(f"Generating new user sheet: '{sheetName}' ")
        spreadSheet = {
                        'properties': {
                            'title': sheetName
                        }
                    }
        spreadSheet = sheetsService.spreadsheets().create(body=spreadSheet, fields='spreadsheetId').execute()
        spreadSheetID = spreadSheet.get('spreadsheetId')
        value_input_option = "USER_ENTERED"
        range = f"Sheet1!A1:D1"
        values = [['Name','Last Name', 'Email', 'Study Year']]
        body = {'values': values }
        sheetsService.spreadsheets().values().update( spreadsheetId=spreadSheetID, range=range, valueInputOption=value_input_option, body=body).execute()
    studentSheetURL = f"https://docs.google.com/spreadsheets/d/{spreadSheetID}/edit#gid=0"
    print(f"studentSheetURL: {studentSheetURL}")
    webbrowser.open_new(studentSheetURL)
    
    if not "studentSheetID" in courseData[courseName]:
        courseData[courseName]["studentSheetID"] = spreadSheetID
        courseData[courseName]["studentSheetURL"] = f"https://docs.google.com/spreadsheets/d/{spreadSheetID}/edit#gid=0"
        with open('Courses.toml', 'w', encoding='utf-8') as file:
            toml.dump(courseData, file)
    
    # Grab spread sheet data
    spreadSheet = sheetsService.spreadsheets().values().get(spreadsheetId=spreadSheetID, range='Sheet1').execute()
    
    courseYears = [year for year,canvasID in courseData[courseName]["canvasIDs"]]
    spreadSheetValues = spreadSheet['values']
    numberOfCurrentEntries = len(spreadSheetValues)
    print(f"Current number of entries: {numberOfCurrentEntries-1}")
    
    # Get student data from Canvas
    print("Looking for unique students in Canvas...")
    studentData = []
    for j, canvasID in enumerate(canvasCourseIDs):
        course = canvas.get_course(canvasID)
        users = course.get_users(enrollment_type=['student'])
        for i,user in enumerate(users):
            name = user.name.split(' ', 1)
            if not hasattr(user, "email"):
                continue
            if not any(user.email in student for student in studentData) and not any(user.email in student for student in spreadSheetValues) :
                studentData.append([name[0], name[1], user.email, courseYears[j]])
                print(f"{name[0]} {name[1]}, {user.email}, {courseYears[j]}")
    
    
    print(f"{len(studentData)} new students found in Canvas.")
    if len(studentData) == 0:
        print("No new students found in Canvas. Nothing is added to the student sheet!")
        return
    
    updateStudentsSheet = messagebox.askyesno("Update student sheet?", f"Do you want to update the existing student sheet with {len(studentData)} new entries?")
    
    # If there is anything new to add, it will be added here
    # Add canvas data to sheet
    if updateStudentsSheet or numberOfCurrentEntries <= 1:
        print("Updating the student sheet...")
        value_input_option = "USER_ENTERED"
        newRange = f"Sheet1!A{numberOfCurrentEntries+1}:D{numberOfCurrentEntries+len(studentData)}"
        print(newRange) 
        body = {'values': studentData }
        sheetsService.spreadsheets().values().update( spreadsheetId=spreadSheetID, range=newRange, valueInputOption=value_input_option, body=body).execute()
        print(f"Student sheet is updated: https://docs.google.com/spreadsheets/d/{spreadSheetID}/edit#gid=0")
    else:
        print("Student sheet not updated.")
        
    return spreadSheetID
    
def GetStudentListFromCanvas(courseName, courseData):
    canvas = Canvas("https://ju.instructure.com/", ReadCanvasToken())
    canvasCourseIDs = courseData[courseName]["canvasCourseIDs"]
    
    # Get student data from Canvas
    courseYears = courseData[courseName]["courseYears"]
    studentData = []
    for j, canvasID in enumerate(canvasCourseIDs):
        course = canvas.get_course(canvasID)
        users = course.get_users(enrollment_type=['student'])
        for i,user in enumerate(users):
            name = user.name.split(' ', 1)
            if not any(user.email in student for student in studentData) :
                studentData.append([name[0], name[1], user.email, courseYears[j]])
                print(f"{name[0]} {name[1]}, {user.email}, {courseYears[j]}")
                
def RemovePassedStudents(courseName, responseSheetID):
    courseData=toml.load("Courses.toml")
    studentSheetID = courseData[courseName]["studentSheetID"]

    sheetsService = build('sheets', 'v4', credentials=AuthenticateAgainstGoogle())
    studentSheet = sheetsService.spreadsheets().values().get(spreadsheetId=studentSheetID, range='Sheet1').execute()
    examSheet = sheetsService.spreadsheets().values().get(spreadsheetId=responseSheetID, range='Sheet1').execute()

    examSheetData = sheetsService.spreadsheets().get(spreadsheetId=responseSheetID).execute()
    
    
    studentSheetValues = studentSheet['values']
    examSheetValues = examSheet['values']

    numberOfPassedStudents = 0
    allIdxToRemove = []
    for i,row in enumerate(examSheetValues):
        if len(row) == 0 or "Day" in row[0] or "Start" in row[0]:
            continue
        grade = row[11]
        email = row[4]
        #print(f"{row[2]} {row[3]}, {email} - {grade}")
        
        if email == "":
            continue
        if grade == "3" or grade == "4" or grade == "5" or grade == "G":
            print(f"{row[2]} {row[3]} has passed the exam and can be removed from the student list.")   
            numberOfPassedStudents += 1  
            idxToRemove = [studentSheetValues.index(student) for student in studentSheetValues if(email.lower().strip() in [prop.lower() for prop in student] )]
            #print(idxToRemove)
            if len(idxToRemove) > 0:
                allIdxToRemove += idxToRemove
    allIdxToRemove = list(set(allIdxToRemove)) # Unique and sorted list
    
    examSheetTitle = examSheetData["properties"]["title"]
    print(f"{numberOfPassedStudents} have passed the exam '{examSheetTitle}'")
    print("Removing passed students from the student list...")
    
    if len(allIdxToRemove) == 0:
        print("Nothing to remove!")
        return

    for index in sorted(allIdxToRemove, reverse=True): # need to delete them in reverse order so that the subsequent indexes dont get thrown off
        print(f"Removing {studentSheetValues[index]}")
        del studentSheetValues[index]



    sheetsService.spreadsheets().values().clear(spreadsheetId=studentSheetID, range='Sheet1').execute()
    sheetsService.spreadsheets().values().update(spreadsheetId=studentSheetID, range='Sheet1', valueInputOption="USER_ENTERED", body={'values': studentSheetValues }).execute()

def disp(GUI, msg):
    print(msg)
    #GUI.console = msg #TODO
    
def TryReadExamConfig(self, config, section, key, supressErrorMsg=False):
    try:
        val = config[section][key]
    except:
        if not supressErrorMsg:
            errorMsg = f"Could not find '{key}' in Exams.toml for '{section}'"
            disp(self, errorMsg)
        val = ""
    return val

def WriteExamData(examData):
    with open('Exams.toml', 'w', encoding='utf-8') as file:
        toml.dump(examData, file)

def WriteCourseData(courseData):
    with open('Courses.toml', 'w', encoding='utf-8') as file:
        toml.dump(courseData, file)

def GetDateOfWeekNumber(targetYear, targetWeek, targetDay=1):
    
    targetYear = int(targetYear)
    targetWeek = int(targetWeek)
    targetDay = int(targetDay)
    
    startDate = datetime.date(targetYear, 1, 1) 
    startDate.isocalendar().week

    day = 0
    delta = datetime.timedelta(days=day)
    endDate = startDate + delta
    week = endDate.isocalendar().week
    year = endDate.isocalendar().year
    weekday = endDate.isocalendar().weekday

    #print(endDate.isocalendar())

    #print(targetWeek, targetYear)
    for day in range(0,365):
        #print(week, year)
        if (week == targetWeek) and (year == targetYear) and (weekday == targetDay):
            break
        day += 1
        delta = datetime.timedelta(days=day)
        endDate = startDate + delta
        #print(endDate.isocalendar())
        week = endDate.isocalendar().week
        year = endDate.isocalendar().year
        weekday = endDate.isocalendar().weekday
    #print(week, day, endDate)
    return endDate

def SendExamSchedule(examID):
    
    course = examID.split('-')[0]
    week = examID.split('-')[1]
    # year = examID.split('-')[2]
    courseData = toml.load("Courses.toml")
    examData = toml.load("Exams.toml")
    courseName = courseData[course]["courseName"]
    
    examSheetID = examData[examID]['examSheetID']
    sheetsService = build('sheets', 'v4', credentials=AuthenticateAgainstGoogle())
    examSheet = sheetsService.spreadsheets().values().get(spreadsheetId=examSheetID, range='Sheet1').execute()
    examSheetValues = examSheet['values']
    
    print(examSheetValues)

    examSheetUri = f"https://docs.google.com/spreadsheets/d/{examSheetID}/edit#gid=0"
    print(f"Exam sheet URI: {examSheetUri}")
    webbrowser.open_new(examSheetUri)
    
    #if not messagebox.askyesno("Send exam schedule to students?", f"Do you want to send the exam schedule to the students?"):
    #    print("Nothing sent!")
    #    return

    
    for i, row in enumerate(examSheetValues):
        if len(row)==0:
            continue
        if row[0]=="Day":
            day = row[1]
            continue
        if row[0]=="Start Time":
            continue
        if len(row)<3:
            continue

        startTime = row[0]
        name = row[2].strip()

        if len(row)<4 or row[3]=='':
             messagebox.showerror("Error", f"Last name missing on row {i+1}!"); return;
        lastname = row[3]

        if len(row)<5 or row[4]=='':
            messagebox.showerror("Error", f"Email is missing on row {i+1}!"); return;
        email = row[4] 
        
        if len(row)<8 or row[7]=='':
            messagebox.showerror("Error",f"Course is missing on row {i+1}!"); return;
        course = row[7]

        if len(row)<10 or row[9]=='':
            messagebox.showerror("Error", f"Exam Room is missing on row {i+1}!"); return;
        room = row[9]
 
        courseName = courseData[course]["courseName"]
        emailSubject = f'{courseName} - upcoming exam schdule'
        
        if len(row) > 10:
            emailSent = row[10]
        else:
            emailSent = ""
        if emailSent != "":
            print(f'Not sending to {name} {lastname}. Already sent on {emailSent}')
            continue
        
        print(f"Sending email to {name} {lastname}...")
        
        if "additionalInfo" in examData[examID]:
            additionalInfo = examData[examID]['additionalInfo']
        else:
            additionalInfo = ""


        
        emailMsg = \
        f'Hello {name} {lastname}!\n\n' \
        f'You are schduled for examination in room {room} on {day}, week {week}. Your examination starts at {startTime}!\n' \
        f'Please be on time!\n\n'\
        f'{additionalInfo}\n\n'\
        f'If you have received this email in error, please notify Mirza.\n\n' \
        f'Best regards,\n' \
        f'Mirza Cenanovic\n\n\n' \
        f'This message was autogenerated.'
        
        print(emailSubject)
        print(emailMsg)
        
        errorSendingEmail = True
        while errorSendingEmail:
            try:
                SendEmail(email, emailSubject, emailMsg)
                errorSendingEmail = False
            except:
                if not messagebox.askyesno("Error sending email?", f"Some Outlook error occured. Make sure Outlook is closed and press yes to try again."):
                    print(f"Email was not sent to {name}.")
                    return
        
        scheduleSentData = datetime.date.today().strftime('%Y-%m-%d')
        sheetsService.spreadsheets().values().update( spreadsheetId=examSheetID, range=f"Sheet1!K{i+1}:K{i+1}" , valueInputOption="USER_ENTERED", body={'values': [[scheduleSentData]] }).execute()

        
def ComposeExamEmail(examID):
    course = examID.split('-')[0]
    courseData = toml.load("Courses.toml")
    examData = toml.load("Exams.toml")
    examFilesURL = examData[examID]["examFilesURL"]
    studentSheetID = courseData[course]['studentSheetID']
    sheetsService = build('sheets', 'v4', credentials=AuthenticateAgainstGoogle())
    studentSheet = sheetsService.spreadsheets().values().get(spreadsheetId=studentSheetID, range='Sheet1').execute()
    studentSheetValues = studentSheet['values']
    courseName = courseData[course]["courseName"]
    courseCode = courseData[course]["courseCode"]
    week = examID.split('-')[1]
    year = examID.split('-')[2]

    emailMsg = \
            f'Greetings!\n\n' \
            f'According to my notes you have not yet passed the course {courseName} - {courseCode}.\n\n' \
            f'You either have labs/ work package left to present or you did not pass the exam.\n' \
            f'If you only have labs/ work packages left to present, please contact us during the next period to arrange for a presentation during work hours.\n\n' \
            f'If you have the examination left to do, then the next examination for the course {courseName} is in week {week} {year}.\n' \
            f'All exam files and instructions can be found here: {examFilesURL}\n' \
            f'Examination will be done orally at JTH and a registration form will be sent out two weeks before the exam.\n' \
            f'At the examination you will have 10 minutes to present your exam problem, solution and analysis using an individual presentation as aid. ' \
            f'After that 10 additional minutes are allotted to questions by the examiners regarding your solution, problem modeling, setup, analysis, and conclusions.\n\n' \
            f'If you have received this email in error, please notify Mirza.\n\n' \
            f'Best regards,\n' \
            f'Mirza Cenanovic\n\n\n' \
            f'This message was autogenerated.'


    recipients = []
    for i, row in enumerate(studentSheetValues):
            if len(row)==0:
                continue
            if row[0]=="Name":
                #print('Name      |  Last name       |  Email                 |  Year  | Hinder | Email Sent')
                #print('------------------------------------------------------------------------------------')
                continue
            name = row[0].strip()
            lastname = row[1].strip()
            email = row[2] 
            studyYear = row[3]  
            
            hinder = "" 
            if len(row) > 4:
                hinder = row[4]
            
            emailPreviouslySent = "" 
            if len(row) > 5:
                emailPreviouslySent = row[5]  
            
            #print(f'{name}   | {lastname}  | {email} | {studyYear} | {hinder} | {emailPreviouslySent}')
            
            if not hinder.lower() == "no":
                print(f"Hinder is not set to 'no', NOT sending email to {name} {lastname}!")
                continue
            if emailPreviouslySent.lower() == 'yes':
                print(f"Email was previously sent to {name} {lastname}!")
                continue

            recipients.append(email)

    recipients = ";".join([str(ele) for ele in recipients])

    outlook = win32.Dispatch('outlook.application')
    mail = outlook.CreateItem(0x0)
    subject = "subject"
    message = "Hello"
    mail.Subject = f'{courseName} - upcoming exam'
    mail.Body = emailMsg
    mail.BCC = recipients
    mail.Display()